@import url('https://fonts.googleapis.com/css2?family=Roboto+Mono:wght@200;300;400;500;600;700&display=swap');

.footnote-reference.brackets::before {content: "[";}
.footnote-reference.brackets::after {content: "]";}
